import uuid

from ..muMLE.api.od import ODAPI
from ..muMLE.concrete_syntax.graphviz.make_url import make_url
from ..muMLE.state.devstate import DevState
from ..muMLE.transformation.rule import RuleMatcherRewriter, PriorityActionGenerator
from ..muMLE.util.module_to_dict import module_to_dict
from ..muMLE.util.simulator import InteractiveDecisionMaker, RandomDecisionMaker
from ..transformation import helpers
from ..muMLE.concrete_syntax.textual_od import renderer
from ...deployment.modelica.exporter import export_modelica

class Simulator:

    def __init__(self, state: DevState, model: uuid.UUID, meta_model: uuid.UUID, ramified_meta_model: uuid.UUID,
                 rules: dict):
        self.state = state
        self.init_model = model
        self.simulation_model = model
        self.meta_model = meta_model
        self.ramified_meta_model = ramified_meta_model
        self.rules = rules
        self.matcher_rewriter = RuleMatcherRewriter(state, meta_model, ramified_meta_model,
                                                    eval_context=self.helper_functions())

    def helper_functions(self):
        return {
            **module_to_dict(helpers),
            'get_od': lambda: ODAPI(self.state, self.simulation_model, self.meta_model)
        }

    def run(self, show_every_step=False):

        for i, (rule_name, rule) in enumerate(self.rules.items()):
            while True:
                result = self.matcher_rewriter.exec_on_first_match(self.simulation_model, rule, rule_name,
                                                                   in_place=True)
                if result is None:
                    break

                self.simulation_model, lhs_match, _ = result
                if show_every_step:
                    print(f"Match: {lhs_match}")
                    self.render_electrical_circuit()

        # self.render_electrical_circuit()
        # self.get_model_od()
        self.to_modelica()

    def get_model_od(self):
        print(renderer.render_od(self.state, self.simulation_model, self.meta_model))

    def to_modelica(self):
        with open("experiment.mo", "w") as f:
            f.write(export_modelica(self.state, self.simulation_model, self.meta_model, "Experiment"))

    def render_electrical_circuit(self):
        od = ODAPI(self.state, self.simulation_model, self.meta_model)

        txt = 'node [style=filled shape=rect fontname="Helvetica,Arial,sans-serif" shape=none height=2 width=2];\n'

        def render_object(name, fillcolor, color, label, shape):
            return f'{name} [fillcolor="{fillcolor}" label=<{label}> color="{color}" shape={shape}]\n'

        def render_edge(src, tgt, color, label):
            return f'{src} -> {tgt} [color="{color}" fontcolor="{color}" label="{label}"];\n'

        for _, system in od.get_all_instances("System", include_subtypes=True):
            name = od.get_name(system)
            txt += render_object(name, "#F8CECC", "#B85450", f"<b>{name}</b>", "hexagon")

        for _, real_connector in od.get_all_instances("RealConnector", include_subtypes=True):
            name = od.get_name(real_connector)
            port = od.get_slot_value(real_connector, "port_name")
            txt += render_object(name, "#DAE8FC", "#6C8EBF", f"<b>{name}</b><br/>Port: {port}", "circle")

        for _, instrument in od.get_all_instances("Instrument", include_subtypes=True):
            name = od.get_name(instrument)
            pattern = od.get_slot_value(instrument, "pattern_uri")
            txt += render_object(name, "#D5E8D4", "#82B366", f"<b>{name}</b><br/>{pattern}", "square")

        for _, wire in od.get_all_instances("Wire", include_subtypes=True):
            name = od.get_name(wire)
            txt += render_object(name, "#FFFFFF", "#FFF", f"<b>{name}</b>", "square")

        for _, instrumentn_connector in od.get_all_instances("InstrumentnConnector", include_subtypes=True):
            name = od.get_name(instrumentn_connector)
            txt += render_object(name, "#E1D5E7", "#9673A6", f"<b>{name}</b>", "diamond")

        for _, conn in od.get_all_instances("ImmutableConnection"):
            name = od.get_name(conn)
            src = od.get_source(conn)
            tgt = od.get_target(conn)
            src_name = od.get_name(src)
            tgt_name = od.get_name(tgt)
            txt += render_edge(src_name, tgt_name, "#CC0000", f"{name}")

        for _, conn in od.get_all_instances("MutableConnection"):
            name = od.get_name(conn)
            src = od.get_source(conn)
            tgt = od.get_target(conn)
            src_name = od.get_name(src)
            tgt_name = od.get_name(tgt)
            txt += render_edge(src_name, tgt_name, "#00CC00", f"{name}")

        for _, conn in od.get_all_instances("InstrumentnPoint"):
            name = od.get_name(conn)
            src = od.get_source(conn)
            tgt = od.get_target(conn)
            src_name = od.get_name(src)
            tgt_name = od.get_name(tgt)
            txt += render_edge(src_name, tgt_name, "#7F00FF", f"{name}")

        for _, conn in od.get_all_instances("AbstractConnection"):
            name = od.get_name(conn)
            src = od.get_source(conn)
            tgt = od.get_target(conn)
            src_name = od.get_name(src)
            tgt_name = od.get_name(tgt)

            if len(od.get_incoming(conn, "HasAbstractConnection")) > 0:
                txt += f'{name}_dummy [shape=point width=0];\n'
                txt += f'{src_name} -> {name}_dummy [color="#0000FF" fontcolor="#0000FF" label="{name}" arrowhead=none];\n'
                txt += f'{name}_dummy -> {tgt_name} [color="#0000FF"];\n'
            else:
                txt += render_edge(src_name, tgt_name, "#0000FF", f"{name}")

        for _, conn in od.get_all_instances("HasAbstractConnection"):
            name = od.get_name(conn)
            src = od.get_source(conn)
            tgt = od.get_target(conn)
            src_name = od.get_name(src)
            tgt_name = f'{od.get_name(tgt)}_dummy'
            txt += render_edge(src_name, tgt_name, "#CCCC00", f"{name}")

        print(make_url(txt))
