:- module(expert_system_server, [
    server/0,
    load_ontology/0
]).

:- use_module(library(http/http_server)).
:- use_module(library(http/thread_httpd)).
:- use_module(library(http/http_dispatch)).
:- use_module(library(http/http_json)).
:- use_module(library(semweb/rdf_db)).
:- use_module(library(semweb/turtle)).
:- use_module(experiment_questions).  % Include the experiment_questions module
:- dynamic current_questions/1, user_answer/2, previous_user_answer/2, triggered_by/2, invalidated_questions/1.

% Define API endpoint handlers
:- http_handler('/api/question', handle_question_request, []). % Fetch the next question type and its options
:- http_handler('/api/answer', handle_answer_request, [method(post)]). % Handle the users answer and provide the next question(s)
:- http_handler('/api/options', handle_options_request, [method(options)]).
:- http_handler('/api/invalidate', handle_invalidate_request, []). % Invalidate irrelevant questions when an answer is deselected
:- http_handler('/api/done', handle_done_request, []).  % Handle the 'done' action


% Initialization: Set the first question
:- retractall(current_questions(_)),
   asserta(current_questions(['Do you already have an experiment specification?'])).

% Initialize invalidated_questions
:- retractall(invalidated_questions(_)),
   asserta(invalidated_questions([])).

% Start server at port 1234
server() :-
    load_ontology,
    http_server(http_dispatch, [port(1234)]).

% Load ontology from TTL files
load_ontology :-
    rdf_load('../knowledge-graph/build/owl/msdl.uantwerpen.be/experiment/description/bundle/classes.ttl'),
    rdf_load('../knowledge-graph/build/owl/msdl.uantwerpen.be/experiment/description/bundle/individuals.ttl'),
    rdf_load('../knowledge-graph/build/owl/msdl.uantwerpen.be/experiment/description/bundle/properties.ttl').

% Enable cross-domain requests
cors_enable(_Request) :-
    format('Access-Control-Allow-Origin: *~n'),
    format('Access-Control-Allow-Headers: Content-Type~n'),
    format('Access-Control-Allow-Methods: GET, POST, OPTIONS~n').

% Handle OPTIONS requests
handle_options_request(_Request) :-
    cors_enable(_Request),
    format('Content-type: text/plain~n~n'),
    format('').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Handle request for the next question(s) %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handle_question_request(Request) :-
    cors_enable(Request), % Enable cross-domain requests
    current_questions(Questions), % Determine current questions
    findall(
        QuestionData,
        (
            member(Question, Questions),
            answer_type_for_question(Question, AnswerType),
            (options_for_question(Question, Options) -> true; Options = []),
            (phase_of_question(Question, Phase) -> true; Phase = -1),
            QuestionData = _{
                question: Question,
                options: Options,
                answerType: AnswerType,
                phase: Phase
            }
        ),
        QuestionsData
    ),
    reply_json(QuestionsData). % Send response as JSON

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Handle the users answer and provide the next question(s). %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handle_answer_request(Request) :-
    cors_enable(Request), % Enable cross-domain requests
    http_read_json_dict(Request, Dict), % Read JSON request into a dictionary
    QAs = Dict.get(answers),  % 'answers' is a list of {question, answer} dictionaries
    
    % Transform each question and answer into strings
    maplist(transform_answer, QAs, TransformedQAs),
    
    store_answers(TransformedQAs),       % Store all answers
    determine_next_questions(TransformedQAs, NextQuestions),  % Determine next questions
    
    % Update current_questions with new questions
    retractall(current_questions(_)),
    asserta(current_questions(NextQuestions)),
    
    % Prepare response
    findall(
        _{
            question: Question,
            options: Options,
            answerType: AnswerType
        },
        (
            member(Question, NextQuestions),
            options_for_question(Question, Options),
            answer_type_for_question(Question, AnswerType)
        ),
        QuestionsData
    ),
    reply_json(QuestionsData).

% Auxiliary function to ensure that question and answer are treated as strings
transform_answer(QA, TransformedQA) :-
    Question = QA.get(question),
    Answer = QA.get(answer),
    
    % Ensure that question and answer are treated as strings
    atom_string(QuestionString, Question),  % Converts the term to string

    (   is_list(Answer)
    ->  maplist(atom_string, Answer, AnswerStrings)
    ;   atom_string(AnswerStrings, Answer) % Converts the term to string
    ),

    % Create the transformed dictionary
    TransformedQA = _{question: QuestionString, answer: AnswerStrings}.

% Store multiple answers
store_answers([]).
store_answers([QA | Rest]) :-
    Question = QA.get(question),
    Answer = QA.get(answer),
    store_answer(Question, Answer),
    store_answers(Rest).

% Determine all next questions based on provided answers
determine_next_questions(QAs, NextQuestions) :-
    findall(
        NextQuestion,
        (
            member(QA, QAs),
            Question = QA.get(question),
            Answer = QA.get(answer),
            next_questions(Question, Answer, NQs),
            member(NextQuestion, NQs)
        ),
        AllNextQuestions
    ),
    list_to_set(AllNextQuestions, NextQuestions).  % Remove duplicates

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Handle question invalidation when an answer is deselected/changed %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handle_invalidate_request(Request) :-
    cors_enable(Request),
    http_read_json_dict(Request, Dict),
    DeselectedAnswer = Dict.get(deselectedAnswer),
    invalidate_dependent_questions(DeselectedAnswer, IrrelevantQuestions),
    reply_json(json([irrelevantQuestions=IrrelevantQuestions])).

% Invalidate dependent questions
invalidate_dependent_questions(DeselectedAnswer, IrrelevantQuestions) :-
    findall(SubQuestion, triggered_by(SubQuestion, DeselectedAnswer), IrrelevantQuestions),
    % Update the list of invalidated questions
    retract(invalidated_questions(CurrentInvalidated)),
    union(CurrentInvalidated, IrrelevantQuestions, NewInvalidated),
    asserta(invalidated_questions(NewInvalidated)).

% Optionally, handle re-validation when an answer is re-selected
handle_revalidate_request(Request) :-
    cors_enable(Request),
    http_read_json_dict(Request, Dict),
    SelectedAnswer = Dict.get(selectedAnswer),
    revalidate_dependent_questions(SelectedAnswer, RelevantQuestions),
    reply_json(json([relevantQuestions=RelevantQuestions])).

revalidate_dependent_questions(SelectedAnswer, RelevantQuestions) :-
    findall(SubQuestion, triggered_by(SubQuestion, SelectedAnswer), RelevantQuestions),
    % Update the list of invalidated questions
    retract(invalidated_questions(CurrentInvalidated)),
    subtract(CurrentInvalidated, RelevantQuestions, NewInvalidated),
    asserta(invalidated_questions(NewInvalidated)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Handle the 'done' action to generate the final file %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handle_done_request(Request) :-
    cors_enable(Request),
    generate_output_file(FilePath),
    reset_state,
    reply_json(json([message='Output file generated', file=FilePath])).

% Reset state to initial configuration
reset_state :-
    % Reset current questions
    retractall(current_questions(_)),
    asserta(current_questions(['Do you already have an experiment specification?'])),
    
    % Clear user answers
    retractall(user_answer(_, _)),
    retractall(previous_user_answer(_, _)),
    
    % Reset invalidated questions
    retractall(invalidated_questions(_)),
    asserta(invalidated_questions([])),

    % Clear triggered and invalidation dependencies (if applicable)
    retractall(triggered_by(_, _)).

% Generate the output file based on user answers
generate_output_file(FilePath) :-
    % For now, just create a placeholder file
    FilePath = './experiment_design_output.txt',
    open(FilePath, write, Stream),
    forall(user_answer(Q, A), format(Stream, '~w: ~w~n', [Q, A])),
    close(Stream).

store_answer(Question, Answer) :-
    % Remove any previous answer for the same question
    retractall(user_answer(Question, _)),
    % Store the new answer
    asserta(user_answer(Question, Answer)).
