:- module(expert_system_cli,
    [
        start_expert_system/0,
        ask_name/1,
        ask_sus/1,
        ask_interfaces/2
    ]).

:- use_module(library(semweb/rdf11)).
:- use_module(library(semweb/turtle)).
:- use_module(library(semweb/rdf_portray)).
:- use_module(library(semweb/rdfs)).
:- use_module(library(janus)).

% Import local modules
:- use_module(questionnaire).
:- use_module(owl_interface).

% Load the RDF/OWL ontology
:- rdf_load('./knowledge-graph/build/owl/msdl.uantwerpen.be/experiment/description/bundle/classes.ttl').
:- rdf_load('./knowledge-graph/build/owl/msdl.uantwerpen.be/experiment/description/bundle/individuals.ttl').
:- rdf_load('./knowledge-graph/build/owl/msdl.uantwerpen.be/experiment/description/bundle/properties.ttl').

% Register URI prefixes to make code readable
:- rdf_register_prefix(experiment, 'http://msdl.uantwerpen.be/experiment/vocabulary/core/experiment#').
:- rdf_register_prefix(instrument, 'http://msdl.uantwerpen.be/experiment/vocabulary/core/instrument#').
:- rdf_register_prefix(poi, 'http://msdl.uantwerpen.be/experiment/vocabulary/core/poi#').
:- rdf_register_prefix(resource, 'http://msdl.uantwerpen.be/experiment/vocabulary/core/resource#').
:- rdf_register_prefix(sus, 'http://msdl.uantwerpen.be/experiment/vocabulary/core/sus#').
:- rdf_register_prefix(exspec, 'http://msdl.uantwerpen.be/experiment/description/experiments/experiment1#').

% Entry point: Start the expert system
start_expert_system :-
    writeln('Starting experiment design expert system ...'),
    ask_name(ExspecName),
    ask_sus(ExspecName),
    ask_time_base(ExspecName),
    ask_atomicity(ExspecName).

% Ask the name of the experiment specificiation
ask_name(ExspecName) :-
    ask_text_question('What is the name of the experiment specification?', ExspecName),
    rdf_assert(exspec:'ExspecName', rdf:type, experiment:'ExperimentSpecification').

% Ask to configure the instrumented sus if a configuration does not already exist
ask_sus(ExspecName) :-
    findall(SuSInstance, rdf(SuSInstance, rdf:type, sus:'System'), SuSList),
    ask_radio_question('What is the SuS?', SuSList, SuS),
    findall(InstrSusInstance, rdf(InstrSusInstance, instrument:isDerivedFromBaseSystem, SuS), InstrSuSList),
    (InstrSuSList \= [] ->
        ask_none_radio_question('Is the instrumented SuS one of these?', InstrSuSList, InstrumentedSuS),
        (InstrumentedSuS == 'None' ->
             ask_text_question('What is the name of the Instrumented SuS?', InstrumentedSuS)
        ;
            !
        )
    ;
        ask_text_question('What is the name of the Instrumented SuS?', InstrumentedSuS)
    ),
    writeln('Asking interfaces ...'),
    rdf_assert(InstrumentedSuS, rdf:type, instrument:'InstrumentedSystem'),
    rdf(InstrumentedSuS, rdf:type, instrument:'InstrumentedSystem'),
    rdf_assert(InstrumentedSuS, instrument:isDerivedFromBaseSystem, SuS),
    rdf_assert(ExspecName, instrument:hasSpecifiedSystemUnderStudy, InstrumentedSuS),
    ask_interfaces(SuS, InstrumentedSuS).

% Ask to configure the interfaces of the instrumented SuS
ask_interfaces(SuS, InstrumentedSuS) :-
    findall(InterfaceInstance, rdf(SuS, instrument:hasInterface, InterfaceInstance), InterfaceList),
    setof(DomainInstance, (member(InterfaceInstance, InterfaceList), rdf(InterfaceInstance, sus:isInDomain, DomainInstance)), DomainList),
    ask_checkbox_question('What are the domains of interfaces associated with the SuS?', DomainList, Domains),
    ask_domain_interfaces(Domains, InstrumentedSuS, SuS).

ask_domain_interfaces([], _, _) :- !.
ask_domain_interfaces([Domain|Rest], InstrumentedSuS, SuS) :-
    findall(InterfaceInstance, (rdf(SuS, instrument:hasInterface, InterfaceInstance), rdf(InterfaceInstance, sus:isInDomain, Domain)), InterfaceList),
    format(atom(Question), 'Select interfaces for domain ~w:', [Domain]),
    ask_checkbox_question(Question, InterfaceList, Interfaces),
    ask_interface_params(Interfaces, InstrumentedSuS),
    ask_domain_interfaces(Rest, InstrumentedSuS, SuS).

ask_interface_params([], _) :- !.
ask_interface_params([Interface|Rest], InstrumentedSuS) :-
    format(atom(Question), 'Is ~w an interface for measurement?', [Interface]),
    ask_boolean_question(Question, IsMeasurement),
    (IsMeasurement == true ->
        rdf_assert(InstrumentedSuS, instrument:hasMeasurementInterface, Interface),
        findall(InterfaceParam, rdf(Interface, sus:hasMeasurementParameter, InterfaceParam), InterfaceParams)
        % reified relation
    ;
        rdf_assert(InstrumentedSuS, instrument:hasStimulusInterface, Interface),
        findall(InterfaceParam, rdf(Interface, sus:hasStimulusParameter, InterfaceParam), InterfaceParams)
        % reified relation
    ),

    ask_interface_params(Rest, InstrumentedSuS).

ask_atomicity(ExspecName, Atomic) :-
    ask_boolean_question('Is the experiment atomic? (select No for composite)', Atomic),
    (Atomic -> change_class(ExspecName, experiment:'AtomicExperimentSpecification'); change_class(ExspecName, experiment:'CompositeExperimentSpecification')).

ask_time_base(ExspecName) :-
    ask_number_question('What is the time-base of the experiment?', TimeBase),
    rdf_assert(ExspecName, experiment:specifiesTimeBase, TimeBase).

% Utility to get a readable label from RDF resources
rdf_label(Resource, Label) :-
    rdf(Resource, rdf:type, _),
    rdf_split_uri(_, LocalExspecName, Resource),
    atom_string(LocalExspecName, Label).
