import { useEffect, useState } from 'react';
import { Button, Typography, Box, List, ListItem, ListItemText, Paper, TextField, Checkbox, FormControlLabel, RadioGroup, Radio, useTheme, FormControl } from '@mui/material';
import ArrowBackIcon from '@mui/icons-material/ArrowBack';
import ArrowForwardIcon from '@mui/icons-material/ArrowForward';
import LoadingWorkflow from './ModalLoading';
import CustomizedSteppers from './Stepper';

type Question = {
  answerType: string;
  options?: string[];
  phase: number;
  question: string;
};

const App = () => {
  const theme = useTheme();
  const [currentQuestionIndex, setCurrentQuestionIndex] = useState(0);
  const [questionary, setQuestionary] = useState<Question[]>([]);
  const [answers, setAnswers] = useState<Record<number, any>>({});
  const [answeredQuestions, setAnsweredQuestions] = useState<Set<number>>(new Set());
  const [isLoading, setIsLoading] = useState(false);
  const [addedPhases, setAddedPhases] = useState<Set<string>>(new Set());

  const fetchQuestions = async () => {
    try {
      const response = await fetch('http://localhost:8081/api/question');
      const newQuestions: Question[] = await response.json();

      setQuestionary((prevQuestionary) => {
        const uniqueQuestions = newQuestions.filter(
          (newQuestion) => !prevQuestionary.some((q) => q.question === newQuestion.question)
        );
        return [...prevQuestionary, ...uniqueQuestions];
      });

      newQuestions.forEach((q) => setAddedPhases((prev) => new Set([...prev, `phase${q.phase}`])));

      if (newQuestions.length === 0) {
        setIsLoading(true);
      }
    } catch (error) {
      console.error('Erro ao buscar perguntas:', error);
    }
  };

  const submitAnswer = async () => {
    const question = questionary[currentQuestionIndex]?.question;
    const answer = answers[currentQuestionIndex];
    const payload = {
      answers: [
        {
          question,
          answer,
        },
      ],
    };

    try {
      const response = await fetch('http://localhost:8081/api/answer', {
        method: 'POST',
        headers: {
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(payload),
      });

      if (!response.ok) {
        throw new Error('Error in post');
      }

      console.log('Answer send with success:', payload);
      await fetchQuestions();
    } catch (error) {
      console.error('Error in post:', error);
    }
  };

  const handleAnswerChange = (answer: any) => {
    setAnswers((prev) => ({
      ...prev,
      [currentQuestionIndex]: answer,
    }));
    setAnsweredQuestions((prev) => new Set([...prev, currentQuestionIndex]));
  };

  const handleNext = () => {
    if (answeredQuestions.has(currentQuestionIndex)) {
      submitAnswer();
      setCurrentQuestionIndex((prev) => Math.min(prev + 1, questionary.length - 1));
    }
  };

  const handlePrev = () => {
    if (currentQuestionIndex > 0) {
      setCurrentQuestionIndex((prev) => prev - 1);
    }
  };

  useEffect(() => {
    fetchQuestions();
  }, []);

  const currentPhase = questionary[currentQuestionIndex]?.phase || 1;

  return (
    <div className="min-h-screen flex flex-col items-center justify-center bg-background-green dark:bg-grey text-text p-6">
      {isLoading ? (
        <LoadingWorkflow />
      ) : (
        <>
          <CustomizedSteppers activeStep={currentPhase - 1} addedPhases={addedPhases} />
          <Paper elevation={3} className="flex w-full mt-12 justify-center max-w-6xl rounded-lg overflow-hidden h-96">
            <Box className="w-1/4 bg-dark-green dark:bg-neutral-900 p-4 overflow-y-auto">
              <Typography variant="h5" fontWeight={700}>
                Phase {currentPhase}
              </Typography>
              <List>
                {questionary.map((_, index) => {
                  const isCurrentQuestion = index === currentQuestionIndex;
                  const isAnswered = answeredQuestions.has(index);

                  return (
                    <ListItem
                      key={index}
                      button
                      selected={isCurrentQuestion}
                      onClick={() => setCurrentQuestionIndex(index)}
                      className={`my-2 rounded-lg ${isAnswered ? 'bg-gradient-to-r from-green-600 via-primary to-primary-light' : 'bg-dark-grey'}`}
                      style={{
                        cursor: 'pointer',
                        border: isCurrentQuestion
                          ? isAnswered
                            ? theme.palette.mode === 'dark'
                              ? '2px solid white'
                              : '2px solid black'
                            : '2px solid var(--neon-green)'
                          : 'none',
                        color:
                          isCurrentQuestion && !isAnswered
                            ? theme.palette.mode === 'dark'
                              ? 'rgb(0,196,167)'
                              : 'white'
                            : 'inherit',
                        background: isAnswered
                          ? theme.palette.mode === 'dark'
                            ? 'linear-gradient(95deg, rgb(0,157,120) 50%, rgb(0,197,167) 100%)'
                            : 'var(--light-green)'
                          : 'inherit',
                      }}
                    >
                      <ListItemText primary={`Question ${index + 1}`} />
                    </ListItem>
                  );
                })}
              </List>
            </Box>

            <Button
              sx={{
                backgroundColor: theme.palette.mode === 'dark' ? 'var(--light-black)' : 'var(--neon-green)',
                color: theme.palette.mode === 'dark' ? 'var(--neon-green)' : 'white',
                borderRadius: 0,
              }}
              onClick={handlePrev}
              startIcon={<ArrowBackIcon />}
              disabled={currentQuestionIndex === 0}
            />

            <Box className="w-3/4 flex flex-col p-8 bg-light-green dark:bg-dark-grey">
              <Typography variant="h5" gutterBottom>
                {questionary[currentQuestionIndex]?.question}
              </Typography>

              {(questionary[currentQuestionIndex]?.answerType === 'text' ||
              questionary[currentQuestionIndex]?.answerType === 'number') && (
                <TextField
                  fullWidth
                  variant="outlined"
                  value={answers[currentQuestionIndex] || ''}
                  onChange={(e) => handleAnswerChange(e.target.value)}
                />
              )}

              {questionary[currentQuestionIndex]?.answerType === 'checkbox' && (
                <FormControl component="fieldset">
                  {questionary[currentQuestionIndex]?.options?.map((option) => (
                    <FormControlLabel
                      key={option}
                      control={
                        <Checkbox
                          checked={answers[currentQuestionIndex]?.includes(option) || false}
                          onChange={(e) => handleAnswerChange(e.target.checked ? [...(answers[currentQuestionIndex] || []), option] : answers[currentQuestionIndex]?.filter((o: string) => o !== option))}
                        />
                      }
                      label={option}
                    />
                  ))}
                </FormControl>
              )}

              {(questionary[currentQuestionIndex]?.answerType === 'radio' || questionary[currentQuestionIndex]?.answerType === 'boolean') && (
                <RadioGroup
                  value={answers[currentQuestionIndex] || ''}
                  onChange={(e) => handleAnswerChange(e.target.value)}
                >
                  {questionary[currentQuestionIndex]?.options?.map((option) => (
                    <FormControlLabel key={option} value={option} control={<Radio />} label={option} />
                  ))}
                </RadioGroup>
              )}
            </Box>

            <Button
              sx={{
                backgroundColor: theme.palette.mode === 'dark' ? 'var(--light-black)' : 'var(--neon-green)',
                color: theme.palette.mode === 'dark' ? 'var(--neon-green)' : 'white',
                borderRadius: 0,
              }}
              onClick={handleNext}
              endIcon={<ArrowForwardIcon />}
              disabled={!answeredQuestions.has(currentQuestionIndex)}
            />
          </Paper>
        </>
      )}
    </div>
  );
};

export default App;
